package org.springframework.cloud.consul.config.watch;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.BeanPostProcessor;
import org.springframework.boot.context.event.ApplicationReadyEvent;
import org.springframework.context.event.EventListener;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.EnableAsync;
import org.springframework.util.StringUtils;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;


/**
 * @author Kysonli
 */

@Slf4j
@EnableAsync
public class ConsulRefreshEventListener implements IConsulRefreshEventListener,BeanPostProcessor {

    private static final String DOT = ".";

    private AtomicBoolean ready = new AtomicBoolean(false);

    private AtomicBoolean init = new AtomicBoolean(false);

    private Map<String,Object> consulData = new LinkedHashMap<String, Object>();

    private Map<String,CallbackItem> configWatchCallbackMap = new HashMap<>();

    @Override
    @EventListener
    public void handle(ApplicationReadyEvent event) {
        this.ready.compareAndSet(false, true);
    }
    @Override
    @EventListener
    public void handle(ConsulRefreshEvent event) {
        Map<String, Object> currentConsulData = event.getRefreshData();
        if (currentConsulData == null) {
            currentConsulData = new LinkedHashMap<>();
        }
        // 初始化
        if (init.compareAndSet(false, true)) {
            consulData = currentConsulData;
            return;
        }
        // Application Ready Event 后才进行回调通知
        if (this.ready.get()) {
            for (Map.Entry<String,Object> entry : currentConsulData.entrySet()) {
                String entryKey = entry.getKey();
                Object currentValue = entry.getValue();
                Object lastValue = consulData.get(entryKey);
                if(equals(lastValue,currentValue)) {
                   continue;
                }
                CallbackItem callbackItem = configWatchCallbackMap.get(entryKey);
                callbackItem = callbackItem == null ? configWatchCallbackMap.get(entryKey.substring(0,entryKey.lastIndexOf(DOT))): callbackItem;
                if (callbackItem == null || callbackItem.callback == null) {
                    continue;
                }
                log.debug("Refresh Consul key changed: " + entryKey + ", lastValue: " + lastValue + ", newValue: " + currentValue);
                if (callbackItem.async) {
                    asyncInvoke(callbackItem.callback,entryKey,lastValue,currentValue);
                } else {
                    syncInvoke(callbackItem.callback,entryKey,lastValue,currentValue);
                }
            }
            consulData = currentConsulData;
        }

    }

    private void doInvoke(ConfigChangeCallback configChangeCallback, String entryKey, Object lastValue, Object currentValue){
        configChangeCallback.callback(new ConfigProperty(entryKey,lastValue),new ConfigProperty(entryKey,currentValue));
    }

    @Async
    public void asyncInvoke(ConfigChangeCallback configChangeCallback, String entryKey, Object lastValue, Object currentValue){
        doInvoke(configChangeCallback,entryKey,lastValue,currentValue);
    }
    @Override
    public void syncInvoke(ConfigChangeCallback configChangeCallback, String entryKey, Object lastValue, Object currentValue){
        doInvoke(configChangeCallback,entryKey,lastValue,currentValue);
    }


    @Override
    public Object postProcessBeforeInitialization(Object o, String s) throws BeansException {
        return o;
    }

    @Override
    public Object postProcessAfterInitialization(Object o, String s) throws BeansException {
        Class<?> clz = o.getClass();
        if (!clz.isAnnotationPresent(ConfigChangeListener.class) || !ConfigChangeCallback.class.isAssignableFrom(clz)) {
            return o;
        }
        ConfigChangeListener configWatchAnnotation = clz.getAnnotation(ConfigChangeListener.class);
        String prefix = configWatchAnnotation.prefix();
        String[] values = configWatchAnnotation.value();
        boolean async = configWatchAnnotation.async();
        if (values.length == 0 && StringUtils.isEmpty(prefix)) {
            return o;
        }
        ConfigChangeCallback bean = (ConfigChangeCallback) o;
        CallbackItem callbackItem = new CallbackItem(bean,async);
        // full match
        if (values.length > 0) {
            for (String value : values){
                configWatchCallbackMap.put(StringUtils.isEmpty(prefix) ? value : prefix + DOT + value,callbackItem);
            }
        } else {
            // prefix match
            configWatchCallbackMap.put(prefix,callbackItem);
        }
        return bean;
    }


    private boolean equals(Object lastValue, Object currentValue) {
        if (lastValue == currentValue) {
            return true;
        }
        if (lastValue == null) {
            return false;
        }
        if(lastValue.equals(currentValue)) {
            return true;
        }
        return false;
    }

    private static class CallbackItem {

        // callback
        public ConfigChangeCallback callback;

        // 回调函数是否需要异步执行
        public boolean async;

        public CallbackItem(ConfigChangeCallback callback, boolean async) {
            this.callback = callback;
            this.async = async;
        }
    }
}
